#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import json
import logging
import os
import typing
from collections import defaultdict
from functools import reduce
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from ai.chronon.api import ttypes
from ai.chronon.group_by import get_output_col_names
from ai.chronon.lineage.lineage_metadata import (
    Config,
    ConfigType,
    LineageMetaData,
    TableType,
)
from ai.chronon.repo.validator import (
    extract_json_confs,
    get_group_by_output_columns,
    get_join_output_columns,
    get_pre_derived_external_features,
    get_pre_derived_group_by_columns,
    get_pre_derived_group_by_features,
    get_pre_derived_join_features,
    get_pre_derived_join_internal_features,
    get_pre_derived_source_keys,
    is_identifier,
)
from ai.chronon.utils import FeatureDisplayKeys, output_table_name, sanitize
from sqlglot import expressions as exp
from sqlglot import maybe_parse
from sqlglot.lineage import Node, to_node
from sqlglot.optimizer import build_scope, normalize_identifiers, qualify

logger = logging.getLogger(__name__)


class LineageParser:
    def __init__(self, schema_provider: Optional[Callable[[str], Dict[str, str]]] = None) -> None:
        """
        Take configurations and generate the following:
         1. Column-level lineage details.
         2. Output tables along with their associated columns.
         3. Feature with corresponding column mappings.

        :param schema_provider: A function that returns a schema dict given a table name.
        """

        self.base_path: Optional[str] = None
        self.team_conf: Optional[Dict[str, Any]] = None
        self.staging_queries: Dict[str, Any] = {}
        self.metadata: LineageMetaData = LineageMetaData()
        self.parsed_staging_query_tables = set()
        self.schema_provider = schema_provider

    def parse_lineage(self, base_path: str, config_filter: Optional[Set[str]] = None) -> LineageMetaData:
        """
        Parse lineage information for staging queries, group bys, and joins using the provided base path.

        :param base_path: Base directory path where configuration files reside.
        :param config_filter: Set of specific config names to process; if None, all are processed.
        :return: The parsed lineage metadata.
        """
        self.base_path = base_path
        teams_config = os.path.join(base_path, "teams.json")
        with open(teams_config) as team_infile:
            self.team_conf = json.load(team_infile)

        # Staging queries are lazily parsed when they are used.
        self.parse_configs(self.handle_staging_query, "production/staging_queries", ttypes.StagingQuery)

        # Parse all group bys and joins.
        self.parse_configs(self.parse_group_by, "production/group_bys", ttypes.GroupBy, config_filter)
        self.parse_configs(self.parse_join, "production/joins", ttypes.Join, config_filter)

        return self.metadata

    def parse_configs(
        self,
        parser: Callable[[Any], None],
        config_path: str,
        config_ttype: ttypes,
        config_filter: Optional[List[str]] = None,
    ) -> None:
        """
        Parse configs from a directory and process each using the provided parser.

        :param parser: Function to process a config.
        :param config_path: Relative path where configs are stored.
        :param config_ttype: Expected type of config objects.
        :param config_filter: Optional list of config names to include.
        """
        path = os.path.join(self.base_path, config_path)
        config_type = config_path.split("/")[-1]
        configs = []
        for root, dirs, files in os.walk(path):
            if root.endswith(config_path):
                continue
            configs.extend(extract_json_confs(config_ttype, root))

        for index, config in enumerate(configs):
            if isinstance(config, config_ttype):
                if config_filter and config.metaData.name not in config_filter:
                    continue
                try:
                    logger.info(f"({index}/{len(configs)}): Parse {config_type} {config.metaData.name} ...")
                    parser(config)
                except Exception as e:
                    self.metadata.unparsed_configs[config_type].append(config.metaData.name)
                    logger.exception(
                        f"An unexpected error occurred while parsing {config_type} {config.metaData.name}: {e}"
                    )

        logger.info(
            f"Total {len(configs)} configs for {config_type}."
            f" Unparsed = {len(self.metadata.unparsed_configs[config_type])}."
        )
        for name in self.metadata.unparsed_configs[config_type]:
            logger.info(f"Unparsed configs: {name}")

    @staticmethod
    def build_select_sql(table: str, selects: List[Tuple[Any, Any]], filter_expr: Optional[Any] = None) -> exp.Select:
        """
        Build a SQL SELECT statement.

        :param table: The table name.
        :param selects: A list of tuples representing select expressions (alias, column).
        :param filter_expr: An optional filter expression for the WHERE clause.
        :return: The built SQLGlot SELECT expression.
        """
        # Clean up any added quotes from select expressions.
        selects = [
            (
                k.replace("`", "") if isinstance(k, str) else k,
                v.replace("`", "") if isinstance(v, str) else v,
            )
            for k, v in selects
        ]
        sql = (
            exp.Select(
                expressions=[exp.alias_(v, k) for k, v in selects],
            )
            .from_(table)
            .where(filter_expr)
        )
        return sql

    def build_aggregate_sql(self, table: str, gb, selects: Optional[List[Tuple[Any, Any]]] = None) -> exp.Select:
        """
        Build a SQL aggregate statement applying AGG() on specified columns.

        :param table: The table name.
        :param gb: The group-by configuration object.
        :param selects: A list of tuples representing select expressions (alias, column).
        :return: The built SQLGlot aggregate SELECT expression.
        """
        expressions = []
        expressions.extend(gb.keyColumns)
        if gb.aggregations:
            for agg in gb.aggregations:
                op_name = ttypes.Operation._VALUES_TO_NAMES[agg.operation]
                agg_names = get_output_col_names(agg)
                for agg_name in agg_names:
                    expressions.append(f"AGG_{op_name}(`{agg.inputColumn}`) AS {agg_name}")
        elif selects:
            for input_column_name, input_column_expr in selects:
                expressions.append(input_column_name)

        expressions.sort()
        sql = exp.Select(expressions=expressions).from_(table)
        return sql

    @staticmethod
    def build_gb_derive_sql(table: str, gb: Any) -> exp.Select:
        """
        Build the SQL derivation query for a group-by configuration.

        :param table: The base table name.
        :param gb: The group-by configuration object.
        :return: The SQLGlot SELECT expression for group-by derivations.
        """
        pre_derived_columns = get_pre_derived_group_by_columns(gb)
        pre_derived_columns.sort()

        output_columns = get_group_by_output_columns(gb)
        output_columns.sort()

        derivation_columns = set(d.name for d in gb.derivations)

        base_sql = exp.Select(expressions=[exp.Column(this=column) for column in pre_derived_columns]).from_(table)

        sql = exp.subquery(base_sql, table).select()

        for derivation in gb.derivations:
            if derivation.name != "*":
                sql = sql.select(exp.alias_(derivation.expression, derivation.name))
        for column in [c for c in output_columns if c not in derivation_columns]:
            sql = sql.select(column)

        return sql

    @staticmethod
    def build_join_derive_sql(table: str, join: Any) -> exp.Select:
        """
        Build the SQL derivation query for a join configuration.

        :param table: The base table name.
        :param join: The join configuration object.
        :return: The SQLGlot SELECT expression for join derivations.
        """
        pre_derived_columns = [value for values in get_pre_derived_join_features(join).values() for value in values]
        pre_derived_columns.extend(get_pre_derived_source_keys(join.left))
        output_columns = get_join_output_columns(join)
        output_columns = (
            output_columns[FeatureDisplayKeys.LEFT_COLUMNS] + output_columns[FeatureDisplayKeys.DERIVED_COLUMNS]
        )

        derivation_columns = set(d.name for d in join.derivations)

        pre_derived_columns.sort()
        output_columns.sort()

        base_sql = exp.Select()
        for column in pre_derived_columns:
            base_sql = base_sql.select(column)
            base_sql = base_sql.from_(table)

        sql = exp.subquery(base_sql, table).select()

        for derivation in join.derivations:
            if derivation.name != "*":
                sql = sql.select(exp.alias_(derivation.expression, derivation.name))
        for column in [c for c in output_columns if c not in derivation_columns]:
            sql = sql.select(column)

        return sql

    @staticmethod
    def base_table_name(table: str) -> str:
        """
        Extract the base table name from a table string with sub partitions.

        :param table: The table string containing sub partitions.
        :return: The base table name.
        """
        return table.split("/")[0]

    def parse_source(self, source: Any) -> Tuple[str, str, Dict[str, str]]:
        """
        Parse the source configuration to extract the table name, filter expression, and select mappings.

        :param source: The source configuration object.
        :return: A tuple containing:
                 - table (str): The table name.
                 - filter_expr (str): The combined filter expression.
                 - selects (Dict[str, str]): Mapping of select expressions.
        """
        if source.entities:
            table = self.base_table_name(source.entities.snapshotTable)
            wheres = source.entities.query.wheres or []
            selects = source.entities.query.selects or {}
        elif source.events:
            table = self.base_table_name(source.events.table)
            wheres = source.events.query.wheres or []
            selects = source.events.query.selects or {}
        else:
            table = self.object_table_name(source.joinSource.join)
            wheres = source.joinSource.query.wheres or []
            selects = source.joinSource.query.selects or {}
            self.parse_join(source.joinSource.join)

        filter_expr = " AND ".join([f"({where})" for where in sorted(wheres)])
        return table, filter_expr, selects

    def get_team(self, metadata: Dict[str, Any]) -> str:
        """
        Extract the team name from the given metadata.

        Retrieves the default team name from the 'team' key, and uses the 'team_override' if available.

        :param metadata: Dictionary containing team information and optional custom JSON.
        :return: The effective team name, possibly overridden by custom configuration.
        """
        # Retrieve the default team name from metadata.
        team = metadata["team"]

        # If 'customJson' is present and contains a 'team_override', use the override.
        if "customJson" in metadata and "team_override" in json.loads(metadata["customJson"]):
            team = json.loads(metadata["customJson"])["team_override"]

        return team

    def object_table_name(self, obj: Any) -> str:
        """
        Construct the fully qualified table name based on the object's team configuration.

        :param obj: The configuration object containing metaData.
        :return: The fully qualified table name.
        """
        if obj.metaData.outputNamespace:
            return output_table_name(obj, full_name=True)
        else:
            sanitized_table_name = sanitize(obj.metaData.name)
            team = self.get_team(obj.metaData)
            if "namespace" in self.team_conf[team]:
                db = self.team_conf[team]["namespace"]
            else:
                db = self.team_conf["default"]["namespace"]
            return db + "." + sanitized_table_name

    @staticmethod
    def check_source_select_non_empty(source: Any) -> Any:
        """
        Check if the source configuration contains non-empty select mappings.

        :param source: The source configuration object.
        :return: The select mappings if present.
        """
        if source.events:
            return source.events.query.selects
        elif source.entities:
            return source.entities.query.selects
        elif source.joinSource:
            return source.joinSource.query.selects

    def build_join_sql(self, join: Any) -> exp.Select:
        """
        Build the SQL LEFT JOIN statement based on the join configuration.

        :param join: The join configuration object.
        :return: The built SQLGlot SELECT expression representing the left join.
        """
        left_table_columns = [value for values in get_pre_derived_join_features(join).values() for value in values]
        left_table_keys = []
        if self.check_source_select_non_empty(join.left):
            left_table_keys = get_pre_derived_source_keys(join.left)
            left_table_columns.extend(left_table_keys)

        join_table = self.object_table_name(join)
        join_table_alias = sanitize(join.metaData.name)
        bootstrap_table = f"{join_table}_bootstrap"

        left_table_columns.sort()
        expressions = []
        for column in left_table_columns:
            if column in left_table_keys:
                expressions.append(exp.Column(this=f"{join_table_alias}.{column}"))
            else:
                expressions.append(exp.Column(this=f"{column}"))

        sql = exp.Select(expressions=expressions).from_(f"{bootstrap_table} AS {join_table_alias}")
        for jp in join.joinParts:
            gb = jp.groupBy
            gb_name = sanitize(gb.metaData.name)
            group_by_columns = get_group_by_output_columns(gb)
            prefix = jp.prefix + "_" if jp.prefix else ""
            join_part_table = f"{prefix}{gb_name}"

            group_by_selects = [(key_column, key_column) for key_column in gb.keyColumns]
            # rename column from join part table to join table by adding the prefix of the join parts
            for column in group_by_columns:
                if column not in gb.keyColumns:
                    group_by_selects.append((f"{join_part_table}_{column}", f"{column}"))

            # sort the column to make the order deterministic
            group_by_selects.sort()

            # select from the join part table
            gb_sql = self.build_select_sql(f"{join_table}_{join_part_table}", group_by_selects)
            conditions = []
            join_expression = []
            for key_column in gb.keyColumns:
                left_key = right_key = key_column
                if jp.keyMapping:
                    for select_key, gp_key in jp.keyMapping.items():
                        if key_column == gp_key:
                            # Verify select_key exists in left table columns since mistypings may occur.
                            if select_key in left_table_columns:
                                left_key = select_key
                                right_key = gp_key
                conditions.append(
                    exp.EQ(
                        this=exp.Column(this=left_key, table=join_table_alias),
                        expression=exp.Column(this=right_key, table=f"gb_{join_part_table}"),
                    )
                )
                join_expression.append(f"({join_table_alias}.{left_key} = gb_{join_part_table}.{right_key})")
            sql = sql.join(
                exp.Subquery(alias=f"gb_{join_part_table}", this=gb_sql),
                on=exp.and_(*conditions),
                join_type="left",
            )
        return sql

    @staticmethod
    def get_join_key_columns(join: Any) -> Set[str]:
        """
        Parse a join configuration and get the key columns for join output tables.

        :param join: The join configuration object.
        :return: Set of key columns.
        """
        key_columns = set()
        for jp in join.joinParts:
            gb = jp.groupBy
            for key_column in gb.keyColumns:
                if jp.keyMapping:
                    for select_key, gp_key in jp.keyMapping.items():
                        if key_column == gp_key:
                            key_columns.add(select_key)
                else:
                    key_columns.add(key_column)
        return key_columns

    def parse_join(self, join: Any) -> None:
        """
        Parse a join configuration and build lineage based on its SQL.

        :param join: The join configuration object.
        :return: None
        """
        # store config
        config_name = join.metaData.name
        config = Config(config_name, ConfigType.JOIN, join)
        self.metadata.configs[config_name] = config

        # source tables used to build lineage
        sources = dict()
        output_table = self.object_table_name(join)

        # build select sql for left
        table, filter_expr, selects = self.parse_source(join.left)
        source_keys = self.get_join_key_columns(join)
        if table in self.staging_queries:
            self.parse_staging_query(self.staging_queries[table])

        # Build lineage for source table --> boostrap table
        bootstrap_selects = [(k, v) for k, v in selects.items()]
        bootstrap_sql = self.build_select_sql(table, bootstrap_selects, filter_expr)
        bootstrap_table = f"{output_table}_bootstrap"
        bootstrap_sql = bootstrap_sql.sql(dialect="spark", pretty=True)
        sources["bootstrap_table"] = bootstrap_sql
        lineages = build_lineage(bootstrap_table, bootstrap_sql)
        self.metadata.store_table(
            config_name,
            bootstrap_table,
            TableType.JOIN_BOOTSTRAP,
            materialized=True,
            key_columns=set(source_keys),
        )
        self.metadata.store_lineage(lineages, output_table)

        for jp in join.joinParts:
            gb = jp.groupBy
            gb_table_name = sanitize(gb.metaData.name)
            prefix = jp.prefix + "_" if jp.prefix else ""
            join_part_table = f"{output_table}_{prefix}{gb_table_name}"
            self.parse_group_by(gb, join_part_table=join_part_table, join_config_name=config_name)

        # Build lineage for join
        join_sql = self.build_join_sql(join)

        sql = join_sql.sql(dialect="spark", pretty=True)
        sources["join_table"] = sql

        features = get_pre_derived_join_internal_features(join)

        # Build derivation SQL if derivations exist.
        if join.derivations:
            derived_features = [derivation.name for derivation in join.derivations]
            # If "*" is defined then append derived features, otherwise only derived features are used.
            if "*" in derived_features:
                derived_features.remove("*")
                features.extend(derived_features)
                # If it is a rename only derivation, then remove the original feature name
                removed_features = [
                    derivation.expression for derivation in join.derivations if is_identifier(derivation.expression)
                ]
                for removed_feature in removed_features:
                    if removed_feature in features:
                        features.remove(removed_feature)
            else:
                features = derived_features

            derive_sql = self.build_join_derive_sql("join_table", join)
            sql = derive_sql.sql(dialect="spark", pretty=True)
            sources["derive_table"] = sql

        # store feature
        for feature in features:
            self.metadata.store_feature(join.metaData.name, ConfigType.JOIN, feature, output_table)
        # no column lineage for external features
        external_features = get_pre_derived_external_features(join)
        for feature in external_features:
            self.metadata.store_feature(join.metaData.name, ConfigType.JOIN, feature)

        # store table
        self.metadata.store_table(
            config_name,
            output_table,
            TableType.JOIN,
            materialized=True,
            key_columns=set(source_keys),
        )

        # store lineage
        parsed_lineages = build_lineage(output_table, sql, sources)
        self.metadata.store_lineage(parsed_lineages, output_table)

    def handle_staging_query(self, staging_query: Any) -> None:
        """
        Store a staging query configuration.

        :param staging_query: The staging query configuration object.
        :return: None
        """
        table_name = output_table_name(staging_query, full_name=True)
        self.staging_queries[table_name] = staging_query

    def parse_staging_query(self, staging_query: Any) -> None:
        """
        Parse a staging query configuration and build its lineage.

        :param query: The staging query configuration object.
        :return: None
        """
        # store config
        config_name = staging_query.metaData.name
        config = Config(config_name, ConfigType.STAGING_QUERY, staging_query)
        self.metadata.configs[config_name] = config

        schema = dict()
        if self.schema_provider:
            tables = self.find_all_tables(staging_query.query)
            for table in tables:
                schema[table] = self.schema_provider(table)

        table_name = output_table_name(staging_query, full_name=True)

        # if we have already parsed this staging query output table then no need to parse again
        if table_name not in self.parsed_staging_query_tables:
            self.parsed_staging_query_tables.add(table_name)
            parsed_lineages = build_lineage(table_name, staging_query.query, schema=schema)
            self.metadata.store_table(config_name, table_name, TableType.STAGING_QUERY)
            self.metadata.store_lineage(parsed_lineages, table_name)

    @staticmethod
    def find_all_tables(sql):
        """
        Find references to physical tables in an expression.

        :param sql: The SQL expression to analyze.
        :return: A set of table names.
        """
        sql = exp.maybe_parse(sql)

        root = build_scope(sql)
        return {
            source.name
            for scope in root.traverse()
            for source in scope.sources.values()
            if isinstance(source, exp.Table)
        }

    def parse_group_by(
        self,
        gb: Any,
        join_part_table: Optional[str] = None,
        join_config_name: Optional[str] = None,
    ) -> None:
        """
        Parse a group-by configuration and build lineage.

        :param gb: The group-by configuration object.
        :param join_part_table: The join part table name for the group by.
        :param join_config_name: The join config name if it is parsing from a join.
        :return: None
        """
        # store config
        config_name = gb.metaData.name
        config = Config(config_name, ConfigType.GROUP_BY, gb)
        self.metadata.configs[config_name] = config

        # source tables used to build lineage
        sources = {}

        # build select sql
        select_sqls = []
        for source in gb.sources:
            table, filter_expr, selects = self.parse_source(source)
            if table in self.staging_queries:
                self.parse_staging_query(self.staging_queries[table])
            self.metadata.store_table(config_name, table, table_type=TableType.OTHER)
            sql = self.build_select_sql(table, selects.items(), filter_expr)
            select_sqls.append(sql)

        select_sql = reduce(exp.union, select_sqls)

        # Build aggregation SQL.
        key_columns = gb.keyColumns
        agg_sql = self.build_aggregate_sql("select_table", gb, selects=selects.items())

        sql = agg_sql.sql(pretty=True)
        sources["select_table"] = select_sql.sql(pretty=True)

        features = get_pre_derived_group_by_features(gb)
        if gb.derivations:
            derived_features = [derivation.name for derivation in gb.derivations]
            # If "*" is defined then append derived features, otherwise only derived features are used.
            if "*" in derived_features:
                derived_features.remove("*")
                features.extend(derived_features)
                # If it is a rename only derivation, then remove the original feature name
                removed_features = [
                    derivation.expression for derivation in gb.derivations if is_identifier(derivation.expression)
                ]
                for removed_feature in removed_features:
                    if removed_feature in features:
                        features.remove(removed_feature)
            else:
                features = derived_features

            derive_sql = self.build_gb_derive_sql("aggregate_table", gb)
            sources["aggregate_table"] = sql
            sql = derive_sql.sql(pretty=True)

        # If join_part_table is defined, then it generates a join part table.
        # If gb.backfill_start_date is defined, then it generates a group by backfill table.
        # If gb.online is True, then it generates an upload table.
        if join_part_table:
            # store table
            config_name = join_config_name or config_name
            self.metadata.store_table(
                config_name,
                join_part_table,
                TableType.JOIN_PART,
                materialized=True,
                key_columns=set(key_columns),
            )

            # store lineage
            parsed_lineages = build_lineage(join_part_table, sql, sources)
            self.metadata.store_lineage(parsed_lineages, join_part_table)
        else:
            if gb.metaData.online:
                output_table = f"{self.object_table_name(gb)}_upload"

                # store table
                self.metadata.store_table(
                    config_name,
                    output_table,
                    TableType.GROUP_BY_UPLOAD,
                    materialized=True,
                    key_columns={"key_json"},
                )

                # Put all key columns into the "key_json" column, and all values into the "value_json" column.
                lineages = build_lineage(output_table, sql, sources)
                updated_lineages = defaultdict(set)
                for (output_table, output_column), lineage in lineages.items():
                    if output_column in key_columns:
                        updated_lineages[(output_table, "key_json")].update(lineage)
                    else:
                        updated_lineages[(output_table, "value_json")].update(lineage)
                self.metadata.store_lineage(updated_lineages, output_table)

            # track feature lineage to the group_by backfill table
            output_table = self.object_table_name(gb)

            # store table
            self.metadata.store_table(
                config_name,
                output_table,
                TableType.GROUP_BY_BACKFILL,
                materialized=gb.backfillStartDate,
                key_columns=set(key_columns),
            )

            # store lineage
            parsed_lineages = build_lineage(output_table, sql, sources)
            self.metadata.store_lineage(parsed_lineages, output_table)

            # store feature
            for feature in features:
                self.metadata.store_feature(gb.metaData.name, ConfigType.GROUP_BY, feature, output_table)


def _get_col_node_name(node: Node, parent_node: Node) -> Tuple[str, str]:
    """
    Get the fully qualified column name from a SQLGlot Node.

    :param node: A SQLGlot Node object representing a column.
    :param node: A SQLGlot Node object representing the parent node.
    :return: The tuple of table name and column name.
    """
    # Use parent.expression if it is a complex type
    if isinstance(parent_node.expression.this, exp.Dot):
        full_name = parent_node.expression.this.sql(pretty=False, identify=False)
        return full_name.split(".")[0], ".".join(full_name.split(".")[1:])
    else:
        column_name = node.name.split(".")[-1]
        # Strip table alias from table expression if present.
        table_exp: exp.Expression = node.expression.copy()
        table_exp.args["alias"] = None
        return table_exp.sql(pretty=False, identify=False), column_name.replace('"', "")


def get_transform(expression: Any) -> Optional[str]:
    """
    Determine the transformation operation name from a SQLGlot expression.

    :param expression: A SQLGlot expression.
    :return: The transformation operation name, or None if the expression is an "Alias" operation.
    """
    if not expression.this:
        return None
    if expression.this.__class__.__name__ == "Anonymous":
        return expression.this.alias_or_name
    elif expression.this.__class__.__name__ == "Column":
        return expression.__class__.__name__ if expression.__class__.__name__ != "Alias" else None
    return expression.this.__class__.__name__


def append_transform(transforms: List[str], transform: str) -> List[str]:
    """
    Append a transform to the list if it is non-empty and not already the last item.

    :param transforms: A list of transform strings.
    :param transform: The transform string to append.
    :return: The updated list of transforms.
    """
    if transform and (not transforms or transforms[-1] != transform):
        return transforms + [transform]
    return transforms


def build_lineage(
    output_table: str,
    sql: str,
    sources: Dict[str, str] = None,
    schema: Dict[str, str] = None,
) -> Dict[Tuple[str, str], Set[Tuple[Tuple[str, str], Tuple[str]]]]:
    """
    Build the lineage mapping from the SQL query and its source queries.

    :param output_table: The fully qualified output table name.
    :param sql: The SQL query string.
    :param sources: Mapping of source names to SQL query strings.
    :param schema: Mapping of source names to source schema.
    :return: A dictionary mapping each output column to a list of input columns and operations to the output column.
    """
    dialect = "spark"
    expression = maybe_parse(sql, dialect=dialect)
    if sources:
        expression = exp.expand(
            expression,
            {k: typing.cast(exp.Query, maybe_parse(v, dialect=dialect)) for k, v in sources.items()},
            dialect=dialect,
        )
    expression = qualify.qualify(
        expression,
        dialect="spark",
        schema=schema,
        **{"validate_qualify_columns": False, "identify": False},
    )
    parsed_lineages = dict()
    scope = build_scope(expression)
    for output_column in [exp.alias_or_name for exp in scope.expression.expressions if exp.alias_or_name != "*"]:
        # Normalize the output column identifier.
        column = normalize_identifiers.normalize_identifiers(output_column, dialect=dialect).name
        column_node = to_node(column, scope, dialect, trim_selects=False)
        input_columns = set()
        transform = get_transform(column_node.expression)
        transforms = [transform] if transform else []
        queue = [(column_node, transforms)]
        while queue:
            parent, transforms = queue.pop(0)
            for child in parent.downstream:
                if isinstance(child.expression, exp.Table):
                    input_columns.add((_get_col_node_name(child, parent), tuple(transforms)))
                child_transform = get_transform(child.expression)
                queue.append((child, append_transform(transforms, child_transform)))
        parsed_lineages[(output_table, output_column)] = input_columns
    return parsed_lineages



"""
Flow is an abstraction of a DAG.
It contains a list of nodes and their dependencies.
It can be visualized as a tree structure.
"""


class Flow:
    def __init__(self, name):
        self.name = name
        self.nodes = set()

    def add_node(self, node):
        self.nodes.add(node)

    def find_node(self, name):
        for node in self.nodes:
            if node.name == name:
                return node
        return None

    def visualize(self, node=None, level=0):
        if node is None:
            starts = [n for n in self.nodes if not any(n in node.dependencies for node in self.nodes)]
            for start in starts:
                self.visualize(start, 0)
            return

        print("    " * level + f"- {node.name}")
        for dependency in node.dependencies:
            self.visualize(dependency, level + 1)



#     Copyright (C) 2025 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
from typing import Dict, List, Optional, Tuple

import ai.chronon.api.ttypes as api
from ai.chronon.join import DataType

logging.basicConfig(level=logging.INFO)

FieldsType = List[Tuple[str, api.TDataType]]


def InferenceSpec(
    model_backend: str,
    model_backend_params: Dict[str, str],
) -> api.InferenceSpec:
    return api.InferenceSpec(
        modelBackend=model_backend,
        modelBackendParams=model_backend_params,
    )


def Model(
    name: str, inference_spec: api.InferenceSpec, input_schema: FieldsType, output_schema: FieldsType
) -> api.Model:
    metadata = api.MetaData(name=name)
    return api.Model(
        metaData=metadata,
        inferenceSpec=inference_spec,
        inputSchema=DataType.STRUCT(f"model_{name}_inputs", *input_schema),
        outputSchema=DataType.STRUCT(f"model_{name}_outputs", *output_schema),
    )


def ModelTransform(
    model: api.Model,
    input_mappings: Optional[Dict[str, str]] = None,
    output_mappings: Optional[Dict[str, str]] = None,
    prefix: Optional[str] = None,
) -> api.ModelTransform:
    return api.ModelTransform(
        model=model,
        inputMappings=input_mappings,
        outputMappings=output_mappings,
        prefix=prefix,
    )


def ModelTransforms(
    transforms: List[api.ModelTransform],
    passthrough_fields: Optional[List[str]] = None,
) -> api.ModelTransforms:
    return api.ModelTransforms(
        transforms=transforms,
        passthroughFields=passthrough_fields,
    )



#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import gc
import importlib
import json
import os
import re
import shutil
import subprocess
import tempfile
from collections.abc import Iterable
from dataclasses import dataclass, fields
from enum import Enum
from typing import Dict, List, Optional, Union, cast

import ai.chronon.api.ttypes as api
import ai.chronon.repo.extract_objects as eo
from ai.chronon.repo import TEAMS_FILE_PATH, teams

ChrononJobTypes = Union[api.GroupBy, api.Join, api.StagingQuery]

chronon_root_path = os.getenv("CHRONON_ROOT", "")  # passed from compile.py


@dataclass
class Modes:
    backfill = "backfill"
    upload = "upload"
    streaming = "streaming"
    stats_summary = "stats-summary"
    label_join = "label-join"
    log_flattener = "log-flattener"
    consistency_metrics_compute = "consistency-metrics-compute"


@dataclass
class SubStage:
    bootstrap = "bootstrap"
    join_parts = "join_parts"


class FeatureDisplayKeys(Enum):
    LEFT_COLUMNS = "left_columns"
    INTERNAL_COLUMNS = "internal_columns"
    EXTERNAL_COLUMNS = "external_columns"
    DERIVED_COLUMNS = "derived_columns"
    OUTPUT_COLUMNS = "output_columns"
    KEY_COLUMNS = "key_columns"


def edit_distance(str1, str2):
    m = len(str1) + 1
    n = len(str2) + 1
    dp = [[0 for _ in range(n)] for _ in range(m)]
    for i in range(m):
        for j in range(n):
            if i == 0:
                dp[i][j] = j
            elif j == 0:
                dp[i][j] = i
            elif str1[i - 1] == str2[j - 1]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = 1 + min(dp[i][j - 1], dp[i - 1][j], dp[i - 1][j - 1])
    return dp[m - 1][n - 1]


class JsonDiffer:
    def __init__(self):
        self.temp_dir = tempfile.mkdtemp()
        self.new_name = "new.json"
        self.old_name = "old.json"

    def diff(self, new_json_str: object, old_json_str: object, skipped_keys=[]) -> str:
        new_json = {
            k: v for k, v in json.loads(new_json_str).items() if k not in skipped_keys
        }
        old_json = {
            k: v for k, v in json.loads(old_json_str).items() if k not in skipped_keys
        }

        with open(os.path.join(self.temp_dir, self.old_name), mode="w") as old, open(
            os.path.join(self.temp_dir, self.new_name), mode="w"
        ) as new:
            old.write(json.dumps(old_json, sort_keys=True, indent=2))
            new.write(json.dumps(new_json, sort_keys=True, indent=2))
        diff_str = subprocess.run(
            ["diff", old.name, new.name], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        return diff_str

    def clean(self):
        shutil.rmtree(self.temp_dir)


def check_contains_single(candidate, valid_items, type_name, name, print_function=repr):
    name_suffix = f"for {name}" if name else ""
    candidate_str = print_function(candidate)
    if not valid_items:
        assert f"{candidate_str}, is not a valid {type_name} because no {type_name}s are specified {name_suffix}"
    elif candidate not in valid_items:
        sorted_items = sorted(
            map(print_function, valid_items),
            key=lambda item: edit_distance(candidate_str, item),
        )
        printed_items = "\n    ".join(sorted_items)
        assert (
            candidate in valid_items
        ), f"""{candidate_str}, is not a valid {type_name} {name_suffix}
Please pick one from:
    {printed_items}
"""


def check_contains(candidates, *args):
    if isinstance(candidates, Iterable) and not isinstance(candidates, str):
        for candidate in candidates:
            check_contains_single(candidate, *args)
    else:
        check_contains_single(candidates, *args)


def get_streaming_sources(group_by: api.GroupBy) -> List[api.Source]:
    """Checks if the group by has a source with streaming enabled."""
    return [source for source in group_by.sources if is_streaming(source)]


def is_streaming(source: api.Source) -> bool:
    """Checks if the source has streaming enabled."""
    return (source.entities and source.entities.mutationTopic is not None) or (
        source.events and source.events.topic is not None
    )


def get_underlying_source(
    source: api.Source,
) -> Union[api.EventSource, api.EntitySource, api.JoinSource]:
    if source.entities:
        return source.entities
    elif source.events:
        return source.events
    else:
        return source.joinSource


def get_query(source: api.Source) -> api.Query:
    return get_underlying_source(source).query


def get_table(source: api.Source) -> str:
    if source.entities:
        table = source.entities.snapshotTable
    elif source.events:
        table = source.events.table
    else:
        table = get_join_output_table_name(source.joinSource.join, True)
    return table.split("/")[0]


def get_topic(source: api.Source) -> str:
    return source.entities.mutationTopic if source.entities else source.events.topic


def get_columns(source: api.Source):
    query = get_query(source)
    assert query.selects is not None, "Please specify selects in your Source/Query"
    columns = query.selects.keys()
    return columns


def get_mod_name_from_gc(obj, mod_prefix):
    """get an object's module information from garbage collector"""
    mod_name = None
    # get obj's module info from garbage collector
    gc.collect()
    for ref in gc.get_referrers(obj):
        if "__name__" in ref and ref["__name__"].startswith(mod_prefix):
            mod_name = ref["__name__"]
            break
    return mod_name


def __set_name(obj, cls, mod_prefix):
    module_name = get_mod_name_from_gc(obj, mod_prefix)
    assert module_name, f"Couldn't find module name for object:\n{obj}\n"
    module = importlib.import_module(module_name)
    eo.import_module_set_name(module, cls)


def sanitize(name):
    """
    From api.Extensions.scala
    Option(name).map(_.replaceAll("[^a-zA-Z0-9_]", "_")).orNull
    """
    if name is not None:
        return re.sub("[^a-zA-Z0-9_]", "_", name)
    return None


def dict_to_bash_commands(d):
    """
    Convert a dict into a bash command substring
    """
    if not d:
        return ""
    bash_commands = []
    for key, value in d.items():
        cmd = (
            f"--{key.replace('_', '-')}={value}"
            if value
            else f"--{key.replace('_', '-')}"
        )
        bash_commands.append(cmd)
    return " ".join(bash_commands)


def dict_to_exports(d):
    if not d:
        return ""
    exports = []
    for key, value in d.items():
        exports.append(f"export {key.upper()}={value}")
    return " && ".join(exports)


def output_table_name(obj, full_name: bool):
    table_name = sanitize(obj.metaData.name)
    db = obj.metaData.outputNamespace
    db = db or "{{ db }}"
    if full_name:
        return db + "." + table_name
    else:
        return table_name


def join_part_name(jp):
    if jp.groupBy is None:
        raise NotImplementedError(
            "Join Part names for non group bys is not implemented."
        )
    if not jp.groupBy.metaData.name and isinstance(jp.groupBy, api.GroupBy):
        __set_name(jp.groupBy, api.GroupBy, "group_bys")
    return "_".join(
        [
            component
            for component in [jp.prefix, sanitize(jp.groupBy.metaData.name)]
            if component is not None
        ]
    )


def join_part_output_table_name(join, jp, full_name: bool = False):
    """
    From api.Extensions.scala

    Join Part output table name.
    To be synced with Scala API.
    def partOutputTable(jp: JoinPart): String = (Seq(join.metaData.outputTable) ++ Option(jp.prefix) :+
      jp.groupBy.metaData.cleanName).mkString("_")
    """
    return "_".join(
        [
            component
            for component in [
                output_table_name(join, full_name),
                join_part_name(jp),
            ]
            if component is not None
        ]
    )


def group_by_output_table_name(obj, full_name: bool = False):
    """
    Group by backfill output table name
    To be synced with api.Extensions.scala
    """
    if not obj.metaData.name:
        __set_name(obj, api.GroupBy, "group_bys")
    return output_table_name(obj, full_name)


def log_table_name(obj, full_name: bool = False):
    return output_table_name(obj, full_name=full_name) + "_logged"


def get_staging_query_output_table_name(
    staging_query: api.StagingQuery, full_name: bool = False
):
    """generate output table name for staging query job"""
    __set_name(staging_query, api.StagingQuery, "staging_queries")
    return output_table_name(staging_query, full_name=full_name)


def get_join_output_table_name(join: api.Join, full_name: bool = False):
    """generate output table name for join backfill job"""
    if isinstance(join, api.Join):
        __set_name(join, api.Join, "joins")
    # set output namespace
    if not join.metaData.outputNamespace:
        team_name = join.metaData.name.split(".")[0]
        namespace = teams.get_team_conf(
            os.path.join(chronon_root_path, TEAMS_FILE_PATH), team_name, "namespace"
        )
        join.metaData.outputNamespace = namespace
    return output_table_name(join, full_name=full_name)


def get_dependencies(
    src: api.Source,
    dependencies: List[str] = None,
    meta_data: api.MetaData = None,
    lag: int = 0,
) -> List[str]:
    query = get_query(src)
    start = query.startPartition
    end = query.endPartition
    if meta_data is not None:
        result = [json.loads(dep) for dep in meta_data.dependencies]
    elif dependencies:
        result = [
            {"name": wait_for_name(dep), "spec": dep, "start": start, "end": end}
            for dep in dependencies
        ]
    else:
        if src.entities and src.entities.mutationTable:
            # Opting to use no lag for all use cases because that the "safe catch-all" case when
            # it comes to dependencies (assuming ds lands before ds + 1). The actual query lag logic
            # is more complicated and depends on temporal/snapshot accuracy for join.
            result = list(
                filter(
                    None,
                    [
                        wait_for_simple_schema(
                            src.entities.snapshotTable, lag, start, end, query=src.entities.query
                        ),
                        wait_for_simple_schema(
                            src.entities.mutationTable, lag, start, end, query=src.entities.query
                        ),
                    ],
                )
            )
        elif src.entities:
            result = [
                wait_for_simple_schema(src.entities.snapshotTable, lag, start, end, query=src.entities.query)
            ]
        elif src.joinSource:
            parentJoinOutputTable = get_join_output_table_name(
                src.joinSource.join, True
            )
            result = [wait_for_simple_schema(parentJoinOutputTable, lag, start, end, query=src.joinSource.query)]
        else:
            result = [wait_for_simple_schema(src.events.table, lag, start, end, query=src.events.query)]
    return [json.dumps(res) for res in result]


def get_bootstrap_dependencies(bootstrap_parts: List[api.BootstrapPart]) -> List[str]:
    if bootstrap_parts is None:
        return []

    dependencies = []
    for bootstrap_part in bootstrap_parts:
        table = bootstrap_part.table
        start = (
            bootstrap_part.query.startPartition
            if bootstrap_part.query is not None
            else None
        )
        end = (
            bootstrap_part.query.endPartition
            if bootstrap_part.query is not None
            else None
        )
        dependencies.append(wait_for_simple_schema(table, 0, start, end, query=bootstrap_part.query))
    return [json.dumps(dep) for dep in dependencies]


def get_label_table_dependencies(label_part) -> List[str]:
    label_info = [
        (label.groupBy.sources, label.groupBy.metaData) for label in label_part.labels
    ]
    label_info = [
        (source, meta_data) for (sources, meta_data) in label_info for source in sources
    ]
    label_dependencies = [
        dep
        for (source, meta_data) in label_info
        for dep in get_dependencies(src=source, meta_data=meta_data)
    ]
    label_dependencies.append(
        json.dumps(
            {
                "name": "wait_for_{{ join_backfill_table }}",
                "spec": "{{ join_backfill_table }}/ds={{ ds }}",
            }
        )
    )
    return label_dependencies


def wait_for_simple_schema(table, lag, start, end, query=None):
    if not table:
        return None
    table_tokens = table.split("/")
    clean_name = table_tokens[0]
    subpartition_spec = "/".join(table_tokens[1:]) if len(table_tokens) > 1 else ""
    partition_col = "ds" if not query else (query.partitionColumn or "ds")
    return {
        "name": "wait_for_{}_{}{}".format(
            clean_name,
            partition_col,
            "" if lag == 0 else f"_minus_{lag}"
        ),
        "spec": "{}/{}={}{}".format(
            clean_name,
            partition_col,
            "{{ ds }}" if lag == 0 else "{{{{ macros.ds_add(ds, -{}) }}}}".format(lag),
            "/{}".format(subpartition_spec) if subpartition_spec else "",
        ),
        "start": start,
        "end": end,
    }


def wait_for_name(dep):
    replace_nonalphanumeric = re.sub("[^a-zA-Z0-9]", "_", dep)
    name = f"wait_for_{replace_nonalphanumeric}"
    return re.sub("_+", "_", name).rstrip("_")


def dedupe_in_order(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def has_topic(group_by: api.GroupBy) -> bool:
    """Find if there's topic or mutationTopic for a source helps define streaming tasks"""
    return any(
        (source.entities and source.entities.mutationTopic)
        or (source.events and source.events.topic)
        for source in group_by.sources
    )


def get_offline_schedule(conf: ChrononJobTypes) -> Optional[str]:
    schedule_interval = conf.metaData.offlineSchedule or "@daily"
    if schedule_interval == "@never":
        return None
    return schedule_interval


def requires_log_flattening_task(conf: ChrononJobTypes) -> bool:
    return (conf.metaData.samplePercent or 0) > 0


def get_modes_tables(conf: ChrononJobTypes) -> Dict[str, List[str]]:
    """Based on a conf get all the applicable modes and tables"""

    if conf.metaData.name:
        table_name = output_table_name(conf, full_name=True)
    else:
        table_name = output_table_name(conf, full_name=False)

    tables = {}

    # from GroupBy
    if isinstance(conf, api.GroupBy):
        group_by = cast(api.GroupBy, conf)
        if group_by.backfillStartDate is not None:
            tables[Modes.backfill] = [f"{table_name}"]

        online = group_by.metaData.online or False
        if online:
            tables[Modes.upload] = [f"{table_name}_upload"]

            temporal_accuracy = group_by.accuracy or False
            streaming = has_topic(group_by)
            if temporal_accuracy or streaming:
                # streaming is a mode that doesn't have a table
                tables[Modes.streaming] = None
    # from Join
    elif isinstance(conf, api.Join):
        join = cast(api.Join, conf)
        if get_offline_schedule(conf) is not None:
            tables[Modes.backfill] = [f"{table_name}"]
            tables[Modes.stats_summary] = [f"{table_name}_daily_stats"]
        if (
            join.metaData.customJson is not None
            and json.loads(join.metaData.customJson).get("check_consistency") is True
        ):
            tables[Modes.consistency_metrics_compute] = [f"{table_name}_consistency"]
        if requires_log_flattening_task(join):
            tables[Modes.log_flattener] = [f"{table_name}_logged"]
        if join.labelPart is not None:
            tables[Modes.label_join] = [
                f"{table_name}_labels",
                f"{table_name}_labeled",
                f"{table_name}_labeled_latest",
            ]

        if conf.bootstrapParts or get_offline_schedule(conf) is not None:
            tables[SubStage.bootstrap] = [f"{table_name}_bootstrap"]

        if conf.joinParts:
            tables[SubStage.join_parts] = [
                join_part_output_table_name(conf, jp) for jp in conf.joinParts
            ]
    # from StagingQuery
    elif isinstance(conf, api.StagingQuery):
        tables[Modes.backfill] = [f"{table_name}"]
    else:
        raise ValueError(f"Unsupported job type {type(conf).__name__}")

    return tables


def get_applicable_modes(conf: ChrononJobTypes) -> List[str]:
    """Based on a conf and mode determine if a conf should define a task."""
    table_map = get_modes_tables(conf)
    modes = [
        x
        for x in table_map.keys()
        if x not in [getattr(SubStage, field.name) for field in fields(SubStage)]
    ]
    return modes


def get_related_table_names(conf: ChrononJobTypes, skip_join_parts=True) -> List[str]:
    """Based on a conf get all the related tables"""

    tables = get_modes_tables(conf)
    related_tables = []  # type: List[str]

    for m in tables:
        if skip_join_parts and m == SubStage.join_parts:
            continue

        # only add modes that have tables (e.g. skips streaming)
        if tables[m]:
            related_tables.extend(tables[m])

    return related_tables


class DotDict(dict):
    def __getattr__(self, attr):
        if attr in self:
            value = self[attr]
            return DotDict(value) if isinstance(value, dict) else value
        return None


def convert_json_to_obj(d):
    if isinstance(d, dict):
        return DotDict({k: convert_json_to_obj(v) for k, v in d.items()})
    elif isinstance(d, list):
        return [convert_json_to_obj(item) for item in d]
    else:
        return d


def get_config_path(join_name: str) -> str:
    assert "." in join_name, f"Invalid join name: {join_name}"
    team_name, config_name = join_name.split(".", 1)
    return f"production/joins/{team_name}/{config_name}"



from datetime import datetime

from ai.chronon.scheduler.interfaces.orchestrator import WorkflowOrchestrator

from airflow import DAG
from airflow.operators.bash_operator import BashOperator


class AirflowAdapter(WorkflowOrchestrator):
    def __init__(self, dag_id, start_date, schedule_interval="@once", airflow_cluster=None):
        self.dag = DAG(
            dag_id,
            start_date=datetime.strptime(start_date, "%Y-%m-%d"),
            schedule_interval=schedule_interval,
        )
        self.airflow_cluster = airflow_cluster

    def setup(self):
        """Initialize a connection to Airflow"""

    def schedule_task(self, node):
        return BashOperator(task_id=node.name, dag=self.dag, bash_command=node.command)

    def set_dependencies(self, task, dependencies):
        task.set_upstream(dependencies)

    def build_dag_from_flow(self, flow):
        node_to_task = {node.name: self.schedule_task(node) for node in flow.nodes}
        for node in flow.nodes:
            task = node_to_task[node.name]
            for dep in node.dependencies:
                dep_task = node_to_task[dep.name]
                self.set_dependencies(task, dep_task)
        return self.dag

    def trigger_run(self):
        """Trigger the DAG run"""



#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple


class TableType(str, Enum):
    """
    Enum representing the various types of tables.
    """

    GROUP_BY_UPLOAD = "group_by_upload"
    GROUP_BY_BACKFILL = "group_by_backfill"
    JOIN_BOOTSTRAP = "join_bootstrap"
    JOIN_PART = "join_part"
    JOIN = "join"
    STAGING_QUERY = "staging_query"
    OTHER = "other"


@dataclass
class Table:
    config_name: str
    table_name: str
    table_type: TableType
    materialized: bool
    key_columns: Set[str] = None
    columns: Set[str] = field(default_factory=set)


class ConfigType(str, Enum):
    """
    Enum representing the various types of config.
    """

    GROUP_BY = "group_by"
    JOIN = "join"
    STAGING_QUERY = "staging_query"


@dataclass(frozen=True)
class Feature:
    config_name: str
    config_type: ConfigType
    feature_name: str
    table_name: Optional[str] = None
    column_name: Optional[str] = None


@dataclass(frozen=True)
class ColumnTransform:
    input_table: str
    input_column: str
    output_table: str
    output_column: str
    transforms: Tuple[str]


@dataclass
class Config:
    config_name: str
    config_type: ConfigType
    t_object: Any
    tables: Dict[str, Table] = field(default_factory=defaultdict)
    features: Dict[str, Feature] = field(default_factory=defaultdict)


class LineageMetaData:
    def __init__(self) -> None:
        """
        Initializes a new instance of LineageMetaData for storing data lineage information.

        """
        self.lineages: Set[ColumnTransform] = set()
        self.tables: Dict[str, Table] = {}
        self.features: Dict[str, Feature] = {}
        self.configs: Dict[str, Config] = {}

        self.unparsed_configs: Dict[str, List[str]] = defaultdict(list)
        self.unparsed_columns: Dict[str, List[str]] = defaultdict(list)

    def store_column(self, table_name: str, column_name: str) -> None:
        """
        Instance method to extract and store a column into its corresponding Table.

        :param table_name: Table name.
        :param column_name: Column name.
        """
        # Create a new Table if one does not exist for the extracted table name.
        if table_name not in self.tables:
            self.tables[table_name] = Table("", table_name, TableType.OTHER, materialized=True)

        # Add the column name to the table's set of columns.
        self.tables[table_name].columns.add(column_name)

    def store_feature(
        self,
        config_name: str,
        config_type: ConfigType,
        feature_name: str,
        output_table: Optional[str] = None,
    ) -> None:
        """
        Instance method to create and store a Feature.

        Builds a feature identifier by combining the entity and feature names.
        If an output table is provided, it constructs a fully qualified column name.

        :param config_name: Name of the config.
        :param config_type: Type of the config - GroupBy or Join.
        :param feature_name: Name of the feature.
        :param output_table: Optional table name where the feature is stored.
        """
        feature_full_name = f"{config_name}.{feature_name}"
        if output_table:
            self.features[feature_full_name] = Feature(
                config_name, config_type, feature_full_name, output_table, feature_name
            )
        else:
            self.features[feature_full_name] = Feature(config_name, config_type, feature_full_name)

        self.configs[config_name].features[feature_full_name] = self.features[feature_full_name]

    def store_table(
        self,
        config_name: str,
        table_name: str,
        table_type: TableType,
        materialized: bool = True,
        key_columns: Set[str] = None,
    ) -> None:
        """
        Instance method to create and store a Table if it does not already exist.

        :param config_name: Name of the config.
        :param table_name: Name of the table.
        :param table_type: The type of the table as defined in TableType.
        :param materialized: If the table is materialized.
        :param key_columns: The key columns used to join with other tables.
        """
        if table_name not in self.tables:
            self.tables[table_name] = Table(config_name, table_name, table_type, materialized, key_columns)
        else:
            # replace existing table with new parsed table type
            existing_table = self.tables[table_name]
            if existing_table.table_type == TableType.OTHER and table_type != TableType.OTHER:
                self.tables[table_name].table_type = table_type
                self.tables[table_name].config_name = config_name
                self.tables[table_name].key_columns = key_columns

        self.configs[config_name].tables[table_name] = self.tables[table_name]

    def store_lineage(
        self,
        parsed_lineages: Dict[Tuple[str, str], Set[Tuple[Tuple[str, str], Tuple[str]]]],
        table_name: str,
    ) -> None:
        """
        Instance method to record lineage information by mapping input columns to output columns with operations.

        For each output column:
          - It stores the output column.
          - If there are no input columns, the output column is recorded as unparsed.
          - Otherwise, for each input column, it stores the input column and records the lineage
            tuple, combining operations (ignoring "Alias" operations).

        :param parsed_lineages: Dictionary mapping output columns to lists of tuples (input_column, operations).
        :param table_name: Table name to associate unparsed columns.
        """
        for (
            output_table,
            output_column,
        ), input_column_transforms in parsed_lineages.items():
            # Store the output column in its table.
            self.store_column(output_table, output_column)

            # If there are no input columns, record the output column as unparsed.
            if not input_column_transforms:
                self.unparsed_columns[table_name].append(output_column)
                continue

            # Process each input column along with its operations.
            for (input_table, input_column), transforms in input_column_transforms:
                self.store_column(input_table, input_column)
                self.lineages.add(
                    ColumnTransform(
                        input_table,
                        input_column,
                        output_table,
                        output_column,
                        # Reverse transform from input_column to output_column
                        transforms[::-1],
                    )
                )

    def filter_lineages(
        self, input_table: Optional[str] = None, output_table: Optional[str] = None
    ) -> Set[ColumnTransform]:
        """
        Instance method to filter stored lineage tuples based on table names.

        Filters the lineage data to return only those tuples where the input column's table
        matches input_table and/or the output column's table matches output_table.

        :param input_table: Optional table name for filtering by input column.
        :param output_table: Optional table name for filtering by output column.
        :return: A set of lineage tuples that match the filter criteria.
        """
        filtered_lineages = self.lineages

        # Filter by input table if provided.
        if input_table:
            filtered_lineages = {lineage for lineage in filtered_lineages if lineage.input_table == input_table}

        # Filter by output table if provided.
        if output_table:
            filtered_lineages = {lineage for lineage in filtered_lineages if lineage.output_table == output_table}

        return filtered_lineages


def extract_table_name(full_column_name: str) -> str:
    """
    Utility function to extract the table name from a fully qualified column name.

    :param full_column_name: Fully qualified column name (e.g., "db.table.column").
    :return: Extracted table name (e.g., "db.table").
    """
    parts = full_column_name.split("./")[0].split(".")
    return ".".join(parts[:-1])


def extract_column_name(full_column_name: str) -> str:
    """
    Utility function to extract the column name from a fully qualified column name.

    :param full_column_name: Fully qualified column name (e.g., "db.table.column").
    :return: Extracted column name (e.g., "column").
    """
    table_name = extract_table_name(full_column_name)
    return full_column_name[len(table_name) + 1 :]



"""Object for checking the downstream dependency of an entity in the Chronon repository.
"""

#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
import logging
import os
from typing import List

from ai.chronon.api.ttypes import GroupBy, Join
from ai.chronon.logger import get_logger
from ai.chronon.repo import GROUP_BY_FOLDER_NAME, JOIN_FOLDER_NAME
from ai.chronon.repo.validator import extract_json_confs


class ChrononEntityDependencyTracker(object):
    def __init__(self, chronon_root_path: str, log_level=logging.INFO):
        self.logger = get_logger(log_level)
        self.chronon_root_path = chronon_root_path

    def extract_conf(self, obj_class: type, path: str) -> object:
        obj = extract_json_confs(obj_class, os.path.join(self.chronon_root_path, path))
        if len(obj) == 0:
            raise Exception(f"No {obj_class} found in {path}")
        elif len(obj) > 1:
            raise Exception(f"Multiple {obj_class} found in {path}")
        return obj[0]

    def get_join_downstream(self, conf_path: str) -> List[object]:
        join = self.extract_conf(Join, conf_path)
        join_name = join.metaData.name
        names_set = set()
        downstream = list()
        group_bys = extract_json_confs(GroupBy, os.path.join(self.chronon_root_path, GROUP_BY_FOLDER_NAME))
        for group_by in group_bys:
            for source in group_by.sources:
                if (
                    source.joinSource
                    and source.joinSource.join.metaData.name == join_name
                    and group_by.metaData.name not in names_set
                ):
                    names_set.add(group_by.metaData.name)
                    downstream.append(group_by)
        return downstream

    def get_group_by_downstream(self, conf_path: str) -> List[object]:
        group_by = self.extract_conf(GroupBy, conf_path)
        group_by_name = group_by.metaData.name
        names_set = set()
        downstream = list()
        joins = extract_json_confs(Join, os.path.join(self.chronon_root_path, JOIN_FOLDER_NAME))
        for join in joins:
            for join_part in join.joinParts:
                if join_part.groupBy.metaData.name == group_by_name and join.metaData.name not in names_set:
                    names_set.add(join.metaData.name)
                    downstream.append(join)
        return downstream

    def get_downstream(self, conf_path: str) -> List[object]:
        if "joins" in conf_path:
            downstream = self.get_join_downstream(conf_path)
        elif "group_bys" in conf_path:
            downstream = self.get_group_by_downstream(conf_path)
        # TODO: Implement the downstream check for staging queries
        elif "staging_queries" in conf_path:
            self.logger.info("Staging queries downstream check not supported yet.")
            downstream = set()
        else:
            raise Exception(f"Invalid conf path: {conf_path}")
        return downstream

    def get_downstream_names(self, conf_path: str) -> List[str]:
        downstream = self.get_downstream(conf_path)
        downstream_name = list()
        for obj in downstream:
            downstream_name.append(obj.metaData.name)
        return downstream_name
